/*
 * Copyright © 2015 Vecly OÜ <info@vecly.com>
 *                  Karel Pärlin
 *                  mart.ruutel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include "cc1111.h"
#include "rf.h"
#include "usb.h"
#include "hal.h"

#include <string.h>

__xdata uint8_t usb_rd_buffer[BUFFER_SIZE+7];

void clock_init(void);
void io_init(void);
void usb_up(void);
void usb_down(void);

void usb_process(void);
void rf_process(void);
void run_bootloader(void);

#define nop() __asm nop __endasm;

__code char FW_NAME[14] = {'m', '4', '3', '3', '-', 'p', 'i', 'c', 'c', 'o', '-', 'f', 'w'};
__code char FW_VERSION[8] = {'0', '.', '2', '-', 'd', 'e', 'v'};

char continuousRssi = 0;

void delay(unsigned char n) {
    unsigned char i = 0, j = 0;
    n <<= 1;
    while(--n != 0)
        while(--i != 0)
          while(--j != 0)
              nop();
}

void main(void) {
    clock_init();
    io_init();
    usb_init();
    init_rf();

    EA = 1;

    usb_up();
    while(1) {
        usb_process();
        rf_process();
    }
}

void usb_process(void) {
    char c;
    __xdata uint16_t mesgLen = 0;    // complete message length
    __xdata uint16_t dataLen;        // data part length indicator
    __xdata uint8_t cmd;    // command
    __xdata uint8_t arg=0;    // command argument
    __xdata uint8_t ci;

    c = usb_pollchar();
    if (c == USB_READ_AGAIN)
        return;
    c = usb_getchar();

    if (c == 0x01) {
        // PLE Terminal messages reading and processing
#ifdef __433MODEM
        PIN6    = 1;
#endif
        usb_rd_buffer[0] = 0x01;
        usb_rd_buffer[1] = usb_getchar();
        dataLen = usb_rd_buffer[1];
        usb_read(&usb_rd_buffer[2], 5 + dataLen + 2 );
        usb_flush();
        // PLE Terminal message
        // length = SOH + LEN + 2xPX + 3xID + LENxDATA + 2xCRC
        mesgLen = 1 + 1 + 2 + 3 + dataLen + 2;
        // Send since after SOH, as SOH is added as a part of TX preamble
#ifdef __433MODEM
        PIN6    = 0;
#endif
        rf_tx(&usb_rd_buffer[1], mesgLen - 1);

    }
    else if (c == 0x11) {
        // Control messages reading and processing
        dataLen = usb_getchar();
        cmd = usb_getchar();
        if( dataLen == 1 )
            arg = usb_getchar();
        usb_flush();
        if (cmd == 0x02) {
            // Launch bootloader
#ifdef __USBEVAL
            LED = 1;
#endif
#ifdef __433MODEM
            PIN5 = 1;
#endif
            // Wait ca 7 sec before reboot. Disconnect client during this time.
            delay(1);
#ifdef __USBEVAL
            LED = 0;
#endif
#ifdef __433MODEM
            PIN5 = 0;
#endif
            run_bootloader();
        } else if (cmd == 0x03) {
            // Firmware name
            usb_putchar(0x11); // DC1
            usb_putchar(sizeof(FW_NAME)); // LENGTH
            usb_putchar(0x03); // COMMAND
            for( ci = 0; ci < sizeof(FW_NAME); ci++ )
                usb_putchar(FW_NAME[ci]);
            usb_flush();
        } else if (cmd == 0x04) {
            // Firmware version
            usb_putchar(0x11); // DC1
            usb_putchar(sizeof(FW_VERSION)); // LENGTH
            usb_putchar(0x04); // COMMAND
            for( ci = 0; ci < sizeof(FW_VERSION); ci++ )
                usb_putchar(FW_VERSION[ci]);
            usb_flush();
        } else if (cmd == 0x06) {
            // Change TX power. Expected values [0..7] (lowest to highest)
            rf_power(arg);
            usb_flush();
        } else if (cmd == 0x07) {
            // Change RF channel. Expected values [1..7]
            rf_ch(arg);
        } else if (cmd == 0x08) {
            // Goto RF Idle mode
            rf_idle();
        } else if (cmd == 0x09) {
            // Activate continuous RX mode
            rf_rx();
        } else if (cmd == 0x0a) {
            // Last RX RSSI request
            if(arg == 0){
                continuousRssi = 0;
            }
            if(arg == 2){
                continuousRssi = 1;
            }
            if(arg == 1){
                usb_putchar(0x11); // DC1
                usb_putchar(1); // LENGTH
                usb_putchar(0x0a); // COMMAND
                usb_putchar(lastRssi);
                usb_flush();
            }

        } else if (cmd == 0x15) {
            if (arg == 0) {
#ifdef __USBEVAL
                LED = 0;
#endif
#ifdef __433MODEM
                PIN5    = 0;
                PIN6    = 0;
                PIN8    = 0;
                PIN10   = 0;
#endif

            }
            if (arg == 1) {
#ifdef __USBEVAL
                LED = 1;
#endif
#ifdef __433MODEM
                PIN5    = 1;
                PIN6    = 1;
                PIN8    = 1;
                PIN10   = 1;
#endif
            }
        } else
            return;

    } else {
        usb_flush();
    }
}

void rf_process(void) {
    if (usb_wr_counter > 0) {
        __xdata uint8_t i = 0;
#ifdef __433MODEM
        PIN8    = 1;
#endif
        // Send RSSI
        if( continuousRssi ){
            usb_putchar(0x11); // DC1
            usb_putchar(1); // LENGTH
            usb_putchar(0x0a); // COMMAND
            usb_putchar(lastRssi);
        }
        // Send data from RX buffer
        while(i < usb_wr_counter) {
            usb_putchar(usb_wr_buffer[i++]);
        }
        usb_flush();
        usb_wr_counter = 0;
#ifdef __433MODEM
        PIN8    = 0;
#endif
    }
}

void run_bootloader(void) {
    if(I2SCLKF0 == 0xf0 && I2SCLKF1 == 0x0d)
        I2SCLKF2 = 0x69;
    // Disable interrupts
    EA = 0;
    IEN0 = IEN1 = IEN2 = 0;
    usb_down();
    SLEEP &= ~SLEEP_USB_EN;
#ifdef __USBEVAL
    LED = 0;
#endif
#ifdef __433MODEM
    PIN5 = 0;
#endif
    __asm
    ljmp 0x00
    __endasm;
}

void clock_init(void) {
    // Switch system clock to crystal oscillator
    CLKCON = (CLKCON & ~CLKCON_OSC_MASK) | (CLKCON_OSC_XTAL);

    while (!(SLEEP & SLEEP_XOSC_STB)) {}

    // Timer tick speed - 187.5 kHz
    // System clock speed - 24 MHz
    CLKCON = ((CLKCON & ~(CLKCON_TICKSPD_MASK | CLKCON_CLKSPD_MASK)) |
      (CLKCON_TICKSPD_1_128 | CLKCON_CLKSPD_1));

    while ((CLKCON & (CLKCON_TICKSPD_MASK|CLKCON_CLKSPD_MASK)) !=
      (CLKCON_TICKSPD_1_128 | CLKCON_CLKSPD_1)
        ) {}
    return;
}

void io_init(void) {
    // LED as output and off
#ifdef __USBEVAL
    P1DIR |= LED_MASK;
    LED = 0;
#endif
    // PINS can be connected to LEDS' with 270 ohm resistor for indication.
    // See C1111 USB dongle reference design.
#ifdef __433MODEM
    P1DIR |= PIN56810_MASK;
    PIN5    = 0;
    PIN6    = 0;
    PIN8    = 0;
    PIN10   = 0;
#endif
}

void usb_up(void) {
    P1DIR |= USB_MASK;
    USB_ENABLE = 1;
}

void usb_down(void) {
    USB_ENABLE = 0;
    P1DIR &= ~USB_MASK;
}
