/*
 * Copyright © 2015 Vecly OÜ <info@vecly.com>
 *                  Karel Pärlin
 *                  mart.ruutel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 */

#include "rf.h"
#include "cc1111.h"
#include "hal.h"

#include <string.h>

static void init_rf_registers(void);

volatile __xdata uint8_t *rf_tx_buffer;
volatile __xdata uint16_t rf_tx_counter = 0;

volatile __xdata uint8_t rf_rx_buffer[BUFFER_SIZE+9];
volatile __xdata uint16_t rf_rx_counter = 0;
volatile __xdata uint16_t rf_rx_packLen = 0;

volatile __xdata uint8_t usb_wr_buffer[BUFFER_SIZE+9];
volatile __xdata uint16_t usb_wr_counter = 0;

volatile __xdata uint8_t rf_status = 0;

volatile __xdata uint8_t lastRssi = 0;

void init_rf(void) {
    // init variables
    rf_status = 0;
    memset(rf_rx_buffer, 0, BUFFER_SIZE);
    rf_rx_counter = 0;
    rf_rx_packLen = 0;

    init_rf_registers();

    // interrupts RFTXRX and RF priority 2
    IP0 = 0;
    IP1 = IP1_IPG0;

    RFTXRXIE = 0;
    RFIM = RFIM_IM_TXUNF | RFIM_IM_RXOVF | RFIM_IM_TIMEOUT | RFIM_IM_DONE | RFIM_IM_SFD;
    RFIF = 0;
    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);
    // IEN2 SFR is not divisible by 8 and therefore each
    // bit is not individually addressable as RFTXRXIE
    IEN2 |= IEN2_RFIE;

    RFST = RFST_SIDLE;
    while (MARCSTATE != RF_MARCSTATE_IDLE)
        ;

    return;
}

static void rf_reset(void) {
    __xdata uint8_t s = MARCSTATE;
    RFST = RFST_SIDLE; while ((MARCSTATE) != RF_MARCSTATE_IDLE);

    switch(s) {
        case RF_MARCSTATE_RX:
            RFST = RFST_SRX;
            break;
        case RF_MARCSTATE_TX:
            RFST = RFST_STX;
            break;
        default:
            // In case reset to IDLE
            return;
    }
    
    while (MARCSTATE == RF_MARCSTATE_IDLE) 
        ;
    return;
}

void rf_tx(__xdata uint8_t *b, uint8_t l) {
    // Keep original PKTLEN
// TODO original PKTLEN keeping is not neccessary, remove it?
    __xdata uint8_t pktlen = PKTLEN;
    PKTLEN = l;

    rf_tx_buffer = b;
    rf_tx_counter = 0;

    // Stay in RX after TX
    // And after RX go to IDLE
    MCSM1 |= RF_MCSM1_TXOFF_MODE_RX;

    RFTXRXIE = 1;
    RFST = RFST_STX;
    while(MARCSTATE != RF_MARCSTATE_TX)
        ;
#ifdef __USBEVAL
    LED = 1;
#endif
#ifdef __433MODEM
    PIN10 = 1;
#endif
    while(MARCSTATE == RF_MARCSTATE_TX)
        ;
#ifdef __USBEVAL
    LED = 0;
#endif
#ifdef __433MODEM
    PIN10 = 0;
#endif
    PKTLEN = pktlen;
    return;
}

void rf_rx(void) {
    memset(rf_rx_buffer, 0, BUFFER_SIZE);
    rf_rx_counter = 0;
    rf_rx_packLen = 0;
    RFTXRXIE = 1;

    S1CON &= ~(S1CON_RFIF_0 | S1CON_RFIF_1);
    RFIF &= ~RFIF_IM_DONE;

    // Stay in RX after RX
    MCSM1 |= RF_MCSM1_RXOFF_MODE_RX;

    RFST = RFST_SRX;
    while(MARCSTATE != RF_MARCSTATE_RX)
        ;
    return;
}

void rf_idle(void) {
    RFTXRXIE = 0;
    MCSM1 = RF_MCSM1_CCA_MODE_RSSI_BELOW_UNLESS_RECEIVING | RF_MCSM1_RXOFF_MODE_IDLE | RF_MCSM1_TXOFF_MODE_IDLE;

    RFST = RFST_SIDLE;
    while(MARCSTATE != RF_MARCSTATE_IDLE)
        ;
    // Switch all LEDs off
#ifdef __433MODEM
    PIN5    = 0;
    PIN6    = 0;
    PIN8    = 0;
    PIN10   = 0;
#endif
    return;
}

void rf_tx_rx_int_handler(void) __interrupt RFTXRX_VECTOR {
    __xdata uint8_t i;

    // Clear interrupt flag
    RFTXRXIF = 0;
    if (MARCSTATE == RF_MARCSTATE_TX) {
        RFD = rf_tx_buffer[rf_tx_counter++];
    } else if (MARCSTATE == RF_MARCSTATE_RX) {
#ifdef __USBEVAL
        LED = 1;
#endif
#ifdef __433MODEM
        PIN5 = 1;
#endif
        // PLE messages detection on RX
        if( rf_rx_counter == 0 ){
            // Restore start SOH, lost by preamble detection
            rf_rx_buffer[rf_rx_counter++] = 0x01;
        }
        if( rf_rx_counter == 3 ){
            // Determine PLE package length
            if( rf_rx_buffer[ 2 ] & 0x02 ){
                // Host to Terminal
                // length = SOH + LEN + 1xPXA + 2xID + DATA[LEN] + 2xCRC
                rf_rx_packLen = 1 + 1 + 1 + 2 + rf_rx_buffer[ 1 ] + 2;
            }
            else {
                // Terminal to Host
                // length = SOH + LEN + 2xPX + 3xID + LENxDATA + 2xCRC
                rf_rx_packLen = 1 + 1 + 2 + 3 + rf_rx_buffer[ 1 ] + 2;
            }
        }
        if(( rf_rx_packLen > 0 ) && ( rf_rx_counter == rf_rx_packLen )){
            // The end of package is reached, end RX
#ifdef __USBEVAL
            LED = 0;
#endif
#ifdef __433MODEM
            PIN5 = 0;
#endif
            lastRssi = RF_RSSI;
            rf_reset();
            // Move data to usb output buffer
            for( i = 0; i < rf_rx_counter; i++ )
                usb_wr_buffer[ i ] = rf_rx_buffer[ i ];
            usb_wr_counter = rf_rx_counter;
            rf_rx_counter = 0;
            rf_rx_packLen = 0;
            return;
        }
        else {
            // Collect received bytes into buffer 
            rf_rx_buffer[rf_rx_counter++] = RFD;
        }
    }
}

void rf_int_handler(void) __interrupt RF_VECTOR {
    // Clear interrupt flags
    S1CON &= (S1CON_RFIF_0 | S1CON_RFIF_1);

    if (RFIF & RFIF_IM_SFD) {
        // Sync word detected
        RFIF &= ~RFIF_IM_SFD;
    } if (RFIF & (RFIF_IM_DONE | RFIF_IM_TIMEOUT)) {
        // Packet received
        RFIF &= ~(RFIF_IM_DONE | RFIF_IM_TIMEOUT);
        rf_status = RFIF_IM_DONE;
    } if (RFIF & RFIF_IM_RXOVF) {
        rf_reset();
        rf_status = RFIF_IM_RXOVF;
        RFIF &= ~RFIF_IM_RXOVF;
    } if (RFIF & RFIF_IM_TXUNF) {
        rf_reset();
        rf_status = RFIF_IM_TXUNF;
        RFIF &= ~RFIF_IM_TXUNF;
    }
}

void rf_ch(uint8_t ch) {
    // Must be done in IDLE
    rf_idle();
    switch (ch)
     {
    case 1:
        RF_CHANNR = 3;
        break;
    case 2:
        RF_CHANNR = 4;
        break;
    case 3:
        RF_CHANNR = 5;
        break;
    case 4:
        RF_CHANNR = 6;
        break;
    case 5:
        RF_CHANNR = 7;
        break;
    case 6:
        RF_CHANNR = 1;
        break;
    case 7:
        RF_CHANNR = 0;
        break;
    default:
        break;
     }
}

void rf_power(uint8_t p) {
    // values 0..7
    RF_FREND0 &= ~RF_FREND0_PA_POWER_MASK;
    RF_FREND0 |= RF_FREND0_PA_POWER_MASK & p;
}

static void init_rf_registers(void) {
    RF_IOCFG2       = 0x00;
    RF_IOCFG1       = 0x00;
    RF_IOCFG0       = 0x00;
    RF_SYNC1        = 0xaa;
    RF_SYNC0        = 0x01;
    RF_PKTLEN       = 0xff;
    RF_PKTCTRL1     = PKTCTRL1_ADR_CHK_NONE;
    RF_PKTCTRL0     = RF_PKTCTRL0_LENGTH_CONFIG_FIXED;
    RF_ADDR         = 0x00;
    RF_CHANNR       = 0x00;
    RF_FSCTRL1      = 0x0c;
    RF_FSCTRL0      = 0x00;
    RF_FREQ2        = 0x12;
    RF_FREQ1        = 0x10;
    RF_FREQ0        = 0x00;
    RF_MDMCFG4      = 0xAA; // 120 kHz RX filter BW
    RF_MDMCFG3      = 0xb5;
    RF_MDMCFG2      = 0x0E;
    RF_MDMCFG1      = 0x42; // Channel spacing 100 kHz
    RF_MDMCFG0      = 0x11;
    RF_DEVIATN      = 0x36; // Deviation 20 kHz
    RF_MCSM2        = 0x07;
    RF_MCSM1        = RF_MCSM1_CCA_MODE_RSSI_BELOW_UNLESS_RECEIVING | RF_MCSM1_RXOFF_MODE_IDLE | RF_MCSM1_TXOFF_MODE_IDLE;
    RF_MCSM0        = RF_MCSM0_FS_AUTOCAL_FROM_IDLE | RF_MCSM0_MAGIC_2;
    RF_FOCCFG       = 0x1F;
    RF_BSCFG        = 0x6C;
    RF_AGCCTRL2     = 0xC3;
    RF_AGCCTRL1     = 0x80;
    RF_AGCCTRL0     = 0xB0;
    RF_FREND1       = 0xb6;
    RF_FREND0       = 0x17; // POWER
    RF_FSCAL3       = 0xe9;
    RF_FSCAL2       = 0x2a;
    RF_FSCAL1       = 0x00;
    RF_FSCAL0       = 0x1f;
    RF_TEST2        = 0x81;
    RF_TEST1        = 0x35;
    RF_TEST0        = 0x09;
    RF_PA_TABLE7    = 0xc0; // 10 dBm
    RF_PA_TABLE6    = 0xc8; // 7 dBm
    RF_PA_TABLE5    = 0x84; // 5 dBm
    RF_PA_TABLE4    = 0x60; // 0 dBm
    RF_PA_TABLE3    = 0x2c; // -5 dBm
    RF_PA_TABLE2    = 0x34; // -10 dBm
    RF_PA_TABLE1    = 0x1d; // -15 dBm
    RF_PA_TABLE0    = 0x0e; // -20 dBm
}
